import functools
import os
import hashlib
import hmac
import math
import struct
from typing import Optional
from functools import reduce
from pyasn1.codec.der import decoder, encoder
from Crypto.Cipher import AES
from Crypto.Hash import HMAC, SHA1, MD5
from .rc4 import RC4
from ..exceptions import KerberosErr
from ..tools import get_random_bytes


def _nfold(ba, nbytes):
    def rotate_right(ba, nbits):
        n = len(ba)
        nbits %= (n * 8)
        if nbits == 0:
            return bytearray(ba)

        val = int.from_bytes(ba, 'big')

        mask = (1 << (n * 8)) - 1
        rotated = ((val >> nbits) | (val << (n * 8 - nbits))) & mask

        return bytearray(rotated.to_bytes(n, 'big'))

    def add_ones_complement(ba1, ba2):
        n = len(ba1)
        res = bytearray(n)
        carry = 0
        for i in range(n-1, -1, -1):
            s = ba1[i] + ba2[i] + carry
            res[i] = s & 0xff
            carry = s >> 8

        while carry:
            for i in range(n-1, -1, -1):
                s = res[i] + carry
                res[i] = s & 0xff
                carry = s >> 8
                if not carry:
                    break
        return res

    slen = len(ba)

    def lcm(a, b):
        return abs(a*b) // math.gcd(a, b)

    lcm_val = lcm(slen, nbytes)
    big_str = bytearray()
    curr = bytearray(ba)
    for _ in range(lcm_val // slen):
        big_str += curr
        curr = rotate_right(curr, 13)

    parts = [big_str[i:i+nbytes] for i in range(0, len(big_str), nbytes)]
    return bytes(reduce(add_ones_complement, parts))


def derive_key(base_key: bytes, usage: int, payload: int) -> bytes:
    key_len = len(base_key)  # 16 or 32

    constant = struct.pack('>I', usage)
    constant += bytes([payload])

    nfolded = _nfold(constant, 16)

    cipher = AES.new(base_key, AES.MODE_ECB)
    b1 = cipher.encrypt(nfolded)

    if key_len == 16:
        return b1

    b2 = cipher.encrypt(b1)
    return b1 + b2


def aes_cts_encrypt(key: bytes, plain: bytes) -> bytes:
    n = len(plain)
    iv = b'\x00' * 16

    if n < 16:
        raise ValueError("Plaintext too short for CTS")

    if n % 16 == 0:
        cipher = AES.new(key, AES.MODE_CBC, iv)
        return cipher.encrypt(plain)

    pad_len = 16 - (n % 16)
    padded_plain = plain + b'\x00' * pad_len

    cipher = AES.new(key, AES.MODE_CBC, iv)
    full_cipher = cipher.encrypt(padded_plain)

    last_block_start = (len(full_cipher) // 16 - 1) * 16
    prev_block_start = last_block_start - 16

    c_all_but_last_two = full_cipher[:prev_block_start]
    c_second_to_last = full_cipher[prev_block_start:last_block_start]
    c_last = full_cipher[last_block_start:]

    return c_all_but_last_two + c_last + c_second_to_last[:n % 16]


def decrypt_kerberos_aes_cts(key: bytes, usage: int, cipher: bytes) -> bytes:
    # Split cipher and HMAC
    ciphertext = cipher[:-12]
    expected_hmac = cipher[-12:]

    # Derive Keys
    ki = derive_key(key, usage, 0x55)
    ke = derive_key(key, usage, 0xAA)

    n = len(ciphertext)
    block_size = 16

    if n < block_size:
        raise ValueError("Ciphertext too short")

    aes_ecb = AES.new(ke, AES.MODE_ECB)

    if n == block_size:
        # Simple one-block case
        plaintext_with_confounder = aes_ecb.decrypt(ciphertext)
    else:
        # NIST CTS Decryption
        m = ((n - block_size - 1) // block_size) * block_size

        p_start = b""
        iv = b'\x00' * block_size
        if m > 0:
            cipher_cbc = AES.new(ke, AES.MODE_CBC, iv)
            p_start = cipher_cbc.decrypt(ciphertext[:m])
            iv = ciphertext[m-block_size:m]

        cn_minus_1 = ciphertext[m: m + block_size]
        cn = ciphertext[m + block_size:]

        tmp = aes_ecb.decrypt(cn_minus_1)

        pn = bytes(a ^ b for a, b in zip(tmp[:len(cn)], cn))

        last_block_full = cn + tmp[len(cn):]

        pn_minus_1 = bytes(a ^ b
                           for a, b
                           in zip(aes_ecb.decrypt(last_block_full), iv))

        plaintext_with_confounder = p_start + pn_minus_1 + pn

    h = HMAC.new(ki, plaintext_with_confounder, SHA1)
    if h.digest()[:12] != expected_hmac:
        raise ValueError(f"HMAC mismatch. Key usage: {usage}")

    return plaintext_with_confounder


def encrypt_kerberos_aes_cts(session_key: bytes,
                             usage: int,
                             plain_text: bytes,
                             confounder: Optional[bytes] = None):
    ki = derive_key(session_key, usage, 0x55)
    ke = derive_key(session_key, usage, 0xAA)

    if confounder is None:
        confounder = get_random_bytes(16)
    plaintext = confounder + plain_text

    h = HMAC.new(ki, plaintext, SHA1)
    checksum = h.digest()[:12]

    aes = AES.new(ke, AES.MODE_CBC, b'\x00' * 16)

    pad_len = (16 - (len(plaintext) % 16)) % 16
    padded_data = plaintext + (b'\x00' * pad_len)

    ctext = aes.encrypt(padded_data)

    last_full_block = ctext[-32:-16]
    truncated_block = ctext[-16:]

    final_ctext = (
        ctext[:-32] +
        truncated_block +
        last_full_block[:16-pad_len]
    )

    return final_ctext + checksum


def encrypt_kerberos_rc4(session_key, usage, plaintext):
    confounder = get_random_bytes(8)
    data_to_encrypt = confounder + plaintext

    usage_bytes = struct.pack('<I', usage)
    k1 = HMAC.new(session_key, usage_bytes, MD5).digest()

    checksum = HMAC.new(k1, data_to_encrypt, MD5).digest()
    k3 = HMAC.new(k1, checksum, MD5).digest()

    cipher = RC4(k3)
    encrypted_data = cipher.encrypt(data_to_encrypt)

    return checksum + encrypted_data


def decrypt_kerberos_rc4(key, usage, data):
    checksum = data[:16]
    encrypted_data = data[16:]

    k1 = hmac.new(key, struct.pack('<I', usage), hashlib.md5).digest()
    k3 = hmac.new(k1, checksum, hashlib.md5).digest()

    cipher = RC4(k3)
    decrypted = cipher.decrypt(encrypted_data)

    return decrypted[8:]  # Skip 8 bytes confounder


def read_session_key(data: bytes) -> bytes:
    payload = data[16:]
    as_rep_part, _ = decoder.decode(payload)
    key_container_bytes = encoder.encode(as_rep_part[0])
    inner_key_seq, _ = decoder.decode(key_container_bytes)

    key_type_bytes = encoder.encode(inner_key_seq[0])
    key_value_bytes = encoder.encode(inner_key_seq[1])

    key_type_obj, _ = decoder.decode(key_type_bytes)
    key_value_obj, _ = decoder.decode(key_value_bytes)

    session_key = bytes(key_value_obj.asOctets())
    _key_type = int(key_type_obj)

    return session_key


KDC_ERR_PREAUTH_REQUIRED = 25

KRB_ERRORS = {
    0:  "KDC_ERR_NONE (No error)",
    1:  "KDC_ERR_NAME_EXP (Client's entry in database has expired)",
    2:  "KDC_ERR_SERVICE_EXP (Server's entry in database has expired)",
    3:  "KDC_ERR_BAD_PVNO (Requested protocol version number not supported)",
    4:  "KDC_ERR_C_OLD_MAST_KVNO (Client's key encrypted in old master key)",
    5:  "KDC_ERR_S_OLD_MAST_KVNO (Server's key encrypted in old master key)",
    6:  "KDC_ERR_C_PRINCIPAL_UNKNOWN (Username not found in Kerberos db)",
    7:  "KDC_ERR_S_PRINCIPAL_UNKNOWN (Server not found; Must use FQDN)",
    8:  "KDC_ERR_PRINCIPAL_NOT_UNIQUE (Multiple principal entries in db)",
    9:  "KDC_ERR_NULL_KEY (The client or server has a null key)",
    10: "KDC_ERR_CANNOT_POSTDATE (Ticket not eligible for postdating)",
    11: "KDC_ERR_NEVER_VALID (Requested starttime is later than end time)",
    12: "KDC_ERR_POLICY (KDC policy rejects request)",
    13: "KDC_ERR_BADOPTION (KDC cannot accommodate requested option)",
    14: "KDC_ERR_ETYPE_NOSUPP (KDC has no support for encryption type)",
    15: "KDC_ERR_SUMTYPE_NOSUPP (KDC has no support for checksum type)",
    16: "KDC_ERR_PADATA_TYPE_NOSUPP (KDC has no support for padata type)",
    17: "KDC_ERR_TRTYPE_NOSUPP (KDC has no support for transited type)",
    18: "KDC_ERR_CLIENT_REVOKED (Clients credentials have been revoked)",
    19: "KDC_ERR_SERVICE_REVOKED (Credentials for server have been revoked)",
    20: "KDC_ERR_TGT_REVOKED (TGT has been revoked)",
    21: "KDC_ERR_CLIENT_NOTYET (Client not yet valid; try again later)",
    22: "KDC_ERR_SERVICE_NOTYET (Server not yet valid; try again later)",
    23: "KDC_ERR_KEY_EXPIRED (Password has expired; change password to reset)",
    24: "KDC_ERR_PREAUTH_FAILED (Pre-auth invalid; check password)",
    25: "KDC_ERR_PREAUTH_REQUIRED (Additional pre-authentication required)",
    26: "KDC_ERR_SERVER_NOMATCH (Requested server and ticket don't match)",
    27: "KDC_ERR_MUST_USE_USER2USER (Server principal valid for usr2usr only)",
    28: "KDC_ERR_PATH_NOT_ACCEPTED (KDC Policy rejects transited path)",
    29: "KDC_ERR_SVC_UNAVAILABLE (A service is not available)",
    31: "KRB_AP_ERR_BAD_INTEGRITY (Integrity check on decrypted field failed)",
    32: "KRB_AP_ERR_TKT_EXPIRED (Ticket expired)",
    33: "KRB_AP_ERR_TKT_NYV (Ticket not yet valid)",
    34: "KRB_AP_ERR_REPEAT (Request is a replay)",
    35: "KRB_AP_ERR_NOT_US (The ticket isn't for us)",
    36: "KRB_AP_ERR_BADMATCH (Ticket/Auth mismatch; check user@domain (FQDN))",
    37: "KRB_AP_ERR_SKEW (Clock skew too great)",
    38: "KRB_AP_ERR_BADADDR (Incorrect net address)",
    39: "KRB_AP_ERR_BADVERSION (Protocol version mismatch)",
    40: "KRB_AP_ERR_MSG_TYPE (Invalid msg type)",
    41: "KRB_AP_ERR_MODIFIED (Message stream modified)",
    42: "KRB_AP_ERR_BADORDER (Message out of order)",
    44: "KRB_AP_ERR_BADKEYVER (Specified version of key is not available)",
    45: "KRB_AP_ERR_NOKEY (Service key not available)",
    46: "KRB_AP_ERR_MUT_FAIL (Mutual authentication failed)",
    47: "KRB_AP_ERR_BADDIRECTION (Incorrect message direction)",
    48: "KRB_AP_ERR_METHOD (Alternative authentication method required)",
    49: "KRB_AP_ERR_BADSEQ (Incorrect sequence number in message)",
    50: "KRB_AP_ERR_INAPP_CKSUM (Inappropriate type of checksum in message)",
    51: "KRB_AP_PATH_NOT_ACCEPTED (Policy rejects transited path)",
    52: "KRB_ERR_RESPONSE_TOO_BIG (Response too big for UDP; retry with TCP)",
    60: "KRB_ERR_GENERIC (Generic error; description in e-text)",
    61: "KRB_ERR_FIELD_TOOLONG (Field is too long for this implementation)",
    62: "KDC_ERROR_CLIENT_NOT_TRUSTED (Reserved for PKINIT)",
    63: "KDC_ERROR_KDC_NOT_TRUSTED (Reserved for PKINIT)",
    64: "KDC_ERROR_INVALID_SIG (Reserved for PKINIT)",
    65: "KDC_ERR_KEY_TOO_WEAK (Reserved for PKINIT)",
    66: "KDC_ERR_CERTIFICATE_MISMATCH (Reserved for PKINIT)",
    67: "KRB_AP_ERR_NO_TGT (No TGT available to validate USER-TO-USER)",
    68: "KDC_ERR_WRONG_REALM (Reserved for future use)",
    69: "KRB_AP_ERR_USER_TO_USER_REQUIRED (Ticket must be for USER-TO-USER)",
    70: "KDC_ERR_CANT_VERIFY_CERTIFICATE (Reserved for PKINIT)",
    71: "KDC_ERR_INVALID_CERTIFICATE (Reserved for PKINIT)",
    72: "KDC_ERR_REVOKED_CERTIFICATE (Reserved for PKINIT)",
    73: "KDC_ERR_REVOCATION_STATUS_UNKNOWN (Reserved for PKINIT)",
    74: "KDC_ERR_REVOCATION_STATUS_UNAVAILABLE (Reserved for PKINIT)",
    75: "KDC_ERR_CLIENT_NAME_MISMATCH (Reserved for PKINIT)",
    76: "KDC_ERR_KDC_NAME_MISMATCH (Reserved for PKINIT)",
    77: (
        "KDC_ERR_INCONSISTENT_KEY_PURPOSE "
        "(Certificate cannot be used for PKINIT client authentication)"),
    78: (
        "KDC_ERR_DIGEST_IN_CERT_NOT_ACCEPTED "
        "(Digest algorithm for the public key is not acceptable)"),
    79: "KDC_ERR_PA_CHECKSUM_MUST_BE_INCLUDED (The paChksum field is missing)",
    80: (
        "KDC_ERR_DIGEST_IN_SIGNED_DATA_NOT_ACCEPTED "
        "(Digest algorithm used by id-pkinit-authData is not acceptable)"),
    81: (
        "KDC_ERR_PUBLIC_KEY_ENCRYPTION_NOT_SUPPORTED "
        "(The KDC does not support public key encryption delivery)"),
    90: "KDC_ERR_PREAUTH_EXPIRED (Pre-authentication has expired)",
    91: "KDC_ERR_MORE_PREAUTH_DATA_REQUIRED (Additional pre-auth required)",
    92: (
        "KDC_ERR_PREAUTH_BAD_AUTHENTICATION_SET "
        "(KDC cannot accommodate requested pre-authentication element)"),
    93: "KDC_ERR_UNKNOWN_CRITICAL_FAST_OPTIONS (Unknown critical option)"
}


def parse_krb_error(data: bytes):
    if data[0] != 0x7e:
        return

    code = -1
    pos = 2 if data[1] < 128 else 2 + (data[1] & 0x7f)
    pos += 2 if data[pos+1] < 128 else 2 + (data[pos+1] & 0x7f)

    while pos < len(data):
        tag = data[pos]
        length = data[pos+1]
        if tag == 0xa6:
            code = data[pos + 4]
            if code == KDC_ERR_PREAUTH_REQUIRED:
                return
            break
        pos += 2 + length

    raise KerberosErr(
        KRB_ERRORS.get(code, f"Unknown Kerberos Error: {code}")
    )


def peel_tag(data: bytes, target_tag: int) -> Optional[bytes]:
    if not data:
        return None
    p = 0
    while p < len(data):
        tag = data[p]
        p += 1
        if p >= len(data):
            break
        lb = data[p]
        p += 1
        if lb & 0x80:
            n_lb = lb & 0x7f
            c_len = int.from_bytes(data[p: p+n_lb], 'big')
            p += n_lb
        else:
            c_len = lb
        if tag == target_tag:
            return data[p: p + c_len]
        p += c_len

    return None



import functools
from typing import Tuple
from .gss import gss_wrap_rc4, gss_wrap_aes, gss_unwrap_rc4, gss_unwrap_aes


def seal_func_kerberos(session_key: bytes, etype: int):
    def _sealer_rc4(
            flags: int,
            seq_num: int,
            message_to_sign: bytes,
            message_to_encrypt: bytes,
            session_key: bytes) -> Tuple[bytes, bytes]:
        return gss_wrap_rc4(session_key,
                            message_to_encrypt,
                            seq_num)

    def _sealer_aes(
            flags: int,
            seq_num: int,
            message_to_sign: bytes,
            message_to_encrypt: bytes,
            session_key: bytes) -> Tuple[bytes, bytes]:
        return gss_wrap_aes(session_key,
                            message_to_encrypt,
                            seq_num)
    if etype in (17, 18):
        sealer = _sealer_aes
    elif etype == 23:
        sealer = _sealer_rc4
    else:
        raise ValueError(f"Invalid E-type: {etype}")
    return functools.partial(sealer, session_key=session_key)


def gss_unwrap_kerberos(session_key: bytes, etype: int):
    def _unwrap_rc4(
            cipher_text: bytes,
            auth_data: bytes,
            session_key: bytes) -> bytes:
        return gss_unwrap_rc4(session_key,
                              cipher_text,
                              auth_data)

    def _unwrap_aes(
            cipher_text: bytes,
            auth_data: bytes,
            session_key: bytes) -> bytes:
        return gss_unwrap_aes(session_key,
                              cipher_text,
                              auth_data)
    if etype in (17, 18):
        unwrapper = _unwrap_aes
    elif etype == 23:
        unwrapper = _unwrap_rc4
    else:
        raise ValueError(f"Invalid E-type: {etype}")
    return functools.partial(unwrapper, session_key=session_key)


def sign_func_kerberos(session_key: bytes, etype: int):
    def _signer_rc4(
            flags: int,
            seq_num: int,
            message_to_sign: bytes,
            session_key: bytes) -> Tuple[bytes, bytes]:
        return gss_wrap_rc4(session_key,
                            message_to_sign,
                            seq_num)

    def _signer_aes(
            flags: int,
            seq_num: int,
            message_to_sign: bytes,
            session_key: bytes) -> Tuple[bytes, bytes]:
        return gss_wrap_aes(session_key,
                            message_to_sign,
                            seq_num)
    if etype in (17, 18):
        signer = _signer_aes
    elif etype == 23:
        signer = _signer_rc4
    else:
        raise ValueError(f"Invalid E-type: {etype}")
    return functools.partial(signer, session_key=session_key)



from typing import Tuple
from Crypto.Protocol.KDF import PBKDF2
from Crypto.Cipher import AES
from Crypto.Hash import SHA1
from .as_req import build_as_req, build_full_as_req
from .kdc import send_kerberos_packet
from .tools import parse_krb_error, peel_tag


def extract_salt_and_etype(error_data: bytes) -> Tuple[str, int]:
    if error_data[0] != 0x7e:
        raise ValueError("Invalid KRB-ERROR packet")

    e_data_idx = error_data.find(b'\xac')
    if e_data_idx == -1:
        raise ValueError("e-data (Tag 12) not found in KRB-ERROR")

    etype_18_marker = b'\x02\x01\x12'
    marker_idx = error_data.find(etype_18_marker, e_data_idx)

    etype = 18
    if marker_idx == -1:
        etype_17_marker = b'\x02\x01\x11'
        marker_idx = error_data.find(etype_17_marker, e_data_idx)
        etype = 17

    if marker_idx == -1:
        raise ValueError("No AES (17/18) etype found in e-data")

    salt_tag_idx = error_data.find(b'\xa1', marker_idx)
    if salt_tag_idx == -1:
        raise ValueError(f"Salt tag (0xa1) not found for etype {etype}")

    salt_start_idx = -1
    for tag in [b'\x1b', b'\x04', b'\x1d']:
        tag_pos = error_data.find(tag, salt_tag_idx, salt_tag_idx + 15)
        if tag_pos != -1:
            salt_start_idx = tag_pos
            break

    if salt_start_idx == -1:
        raise ValueError(
            "Could not find string tag (1b/04) within salt container")

    pos = salt_start_idx + 1
    length_byte = error_data[pos]
    pos += 1

    if length_byte & 0x80:
        n_bytes = length_byte & 0x7f
        salt_len = int.from_bytes(error_data[pos: pos + n_bytes], 'big')
        pos += n_bytes
    else:
        salt_len = length_byte

    salt_bytes = error_data[pos: pos + salt_len]
    if not salt_bytes:
        raise ValueError("Extracted salt is empty")

    salt = salt_bytes.decode('utf-8', errors='replace')
    return salt, etype


def nfold_for_kerberos():
    return b'\x6b\x65\x72\x62\x65\x72\x6f\x73\x7b\x9b\x5b\x2b\x93\x13\x2b\x93'


def aes_string_to_key(password: str, salt: str, key_len: int = 32):
    """
    Kerberos String-to-Key for AES (RFC 3962)
    key_len should be 16 for AES-128 or 32 for AES-256
    """
    tkey = PBKDF2(
        password,
        salt.encode(),
        dkLen=key_len,
        count=4096,
        hmac_hash_module=SHA1
    )

    constant = nfold_for_kerberos()

    # Kerberos DK (Random-to-Key) logic
    iv = b'\x00' * 16

    cipher1 = AES.new(tkey, AES.MODE_CBC, iv)
    part1 = cipher1.encrypt(constant)

    if key_len == 16:
        # For AES-128, the key is just the first 16-byte block
        return part1
    else:
        # For AES-256, we need a second 16-byte block
        cipher2 = AES.new(tkey, AES.MODE_CBC, iv)
        part2 = cipher2.encrypt(part1)
        return part1 + part2


async def get_tgt(username: str, password: str, domain: str,
                  kdc_host: str, kdc_port: int = 88) -> Tuple[bytes, bytes]:
    as_req = build_as_req(username, domain)
    resp = await send_kerberos_packet(as_req, kdc_host, kdc_port)
    parse_krb_error(resp)
    salt, etype = extract_salt_and_etype(resp)
    if etype != 18:
        raise ValueError('Only AES256 encryption supported for negotiation')
    base_key = aes_string_to_key(password, salt)
    full_as_req = build_full_as_req(username, domain, base_key, etype)
    as_res_bytes = await send_kerberos_packet(full_as_req, kdc_host, kdc_port)
    parse_krb_error(as_res_bytes)
    return as_res_bytes, base_key



import random
from typing import Tuple
from Crypto.Cipher import AES
from datetime import datetime, timezone, timedelta
from pyasn1.codec.der import decoder, encoder
from pyasn1.type import univ, tag, namedtype, char
from .asn1 import (
    asn1_len, asn1_tag, asn1_seq, asn1_gs, asn1_ostr, asn1_int, asn1_gt
)
from .kdc import send_kerberos_packet
from .tools import decrypt_kerberos_aes_cts, encrypt_kerberos_aes_cts
from .tools import read_session_key, parse_krb_error
from .asn1 import get_asn1_len


def aes_cts_encrypt(key, plaintext):
    iv = b'\x00' * 16
    n = len(plaintext)

    if n % 16 == 0:
        cipher = AES.new(key, AES.MODE_CBC, iv)
        return cipher.encrypt(plaintext)

    if n < 16:
        raise ValueError("Plaintext must be at least 16 bytes for AES-CTS")

    pad_len = 16 - (n % 16)
    padded_plain = plaintext + b'\x00' * pad_len

    cipher = AES.new(key, AES.MODE_CBC, iv)
    full_cipher = cipher.encrypt(padded_plain)

    last_block_start = (len(full_cipher) // 16 - 1) * 16
    prev_block_start = last_block_start - 16

    c_prev = full_cipher[prev_block_start:last_block_start]
    c_last = full_cipher[last_block_start:]

    return full_cipher[:prev_block_start] + c_last + c_prev[:n % 16]


class AS_REP(univ.Sequence):
    tagSet = univ.Sequence.tagSet.tagExplicitly(
        tag.Tag(tag.tagClassApplication, tag.tagFormatConstructed, 11)
    )

    componentType = namedtype.NamedTypes(
        namedtype.NamedType('pvno', univ.Integer().subtype(
            explicitTag=tag.Tag(tag.tagClassContext,
                                tag.tagFormatConstructed, 0))),
        namedtype.NamedType('msg-type', univ.Integer().subtype(
            explicitTag=tag.Tag(tag.tagClassContext,
                                tag.tagFormatConstructed, 1))),
        namedtype.OptionalNamedType(
            'padata',
            univ.SequenceOf(univ.Sequence()).subtype(
                explicitTag=tag.Tag(tag.tagClassContext,
                                    tag.tagFormatConstructed, 2))),
        namedtype.NamedType('crealm', char.GeneralString().subtype(
            explicitTag=tag.Tag(tag.tagClassContext,
                                tag.tagFormatConstructed, 3))),
        namedtype.NamedType('cname', univ.Any().subtype(
            explicitTag=tag.Tag(tag.tagClassContext,
                                tag.tagFormatConstructed, 4))),
        namedtype.NamedType('ticket', univ.Any().subtype(
            explicitTag=tag.Tag(tag.tagClassContext,
                                tag.tagFormatConstructed, 5))),
        namedtype.NamedType('enc-part', univ.Any().subtype(
            explicitTag=tag.Tag(tag.tagClassContext,
                                tag.tagFormatConstructed, 6)))
    )


class EncryptedData(univ.Sequence):
    componentType = namedtype.NamedTypes(
        namedtype.NamedType('etype', univ.Integer().subtype(
            explicitTag=tag.Tag(tag.tagClassContext,
                                tag.tagFormatConstructed, 0))),
        namedtype.OptionalNamedType('kvno', univ.Integer().subtype(
            explicitTag=tag.Tag(tag.tagClassContext,
                                tag.tagFormatConstructed, 1))),
        namedtype.NamedType('cipher', univ.OctetString().subtype(
            explicitTag=tag.Tag(tag.tagClassContext,
                                tag.tagFormatConstructed, 2)))
    )


def get_session_key(as_rep_bytes: bytes, base_key: bytes) -> bytes:
    as_rep_obj, _ = decoder.decode(as_rep_bytes, asn1Spec=AS_REP())

    enc_part_any = as_rep_obj['enc-part']
    enc_data_obj, _ = decoder.decode(
        bytes(enc_part_any),
        asn1Spec=EncryptedData()
    )

    cipher_blob = bytes(enc_data_obj['cipher'])

    decrypted = decrypt_kerberos_aes_cts(base_key, 3, cipher_blob)
    session_key = read_session_key(decrypted)

    return session_key


def build_tgs_req(username: str,
                  domain: str,
                  session_key: bytes,
                  ticket_bytes: bytes,
                  target_service: Tuple[str, str]):
    now = datetime.now(timezone.utc)
    ts_str = now.strftime("%Y%m%d%H%M%SZ").encode()
    till_ts = (now + timedelta(hours=8)).strftime("%Y%m%d%H%M%SZ").encode()

    cname_content = (
        asn1_tag(0, asn1_int(1)) +
        asn1_tag(1, asn1_seq(asn1_gs(username.encode())))
    )

    domcaps = domain.upper()
    auth_body = (
        asn1_tag(0, asn1_int(5)) +                       # pvno
        asn1_tag(1, asn1_gs(domcaps.encode())) +         # crealm
        asn1_tag(2, asn1_seq(cname_content)) +           # cname
        asn1_tag(4, asn1_int(now.microsecond)) +         # cusec
        asn1_tag(5, asn1_gt(ts_str))                     # ctime
    )

    # Application Tag 2 (0x62) -> Authenticator
    auth_plain = b'\x62' + asn1_len(asn1_seq(auth_body)) + asn1_seq(auth_body)

    # Encrypt Authenticator (7 -> TGS-REQ) ---
    final_cipher = encrypt_kerberos_aes_cts(session_key, 7, auth_plain)
    auth_enc_seq = asn1_seq(
        asn1_tag(0, asn1_int(18)) +                      # etype AES256
        asn1_tag(2, asn1_ostr(final_cipher))               # cipher
    )

    # AP-REQ (-> PA-DATA)
    ap_req_body = (
        asn1_tag(0, asn1_int(5)) +                       # pvno
        asn1_tag(1, asn1_int(14)) +                      # msg-type (14=AP-REQ)
        asn1_tag(2, b'\x03\x05\x00\x00\x00\x00\x00') +   # ap-options
        asn1_tag(3, ticket_bytes) +                      # ticket
        asn1_tag(4, auth_enc_seq)                        # authenticator
    )
    ap_req_seq = asn1_seq(ap_req_body)
    encoded_ap_req = b'\x6e' + asn1_len(ap_req_seq) + ap_req_seq

    padata_item = asn1_seq(
        asn1_tag(1, asn1_int(1)) +                       # PA-TGS-REQ
        asn1_tag(2, asn1_ostr(encoded_ap_req))           # value
    )
    padata_field = asn1_tag(3, asn1_seq(padata_item))    # [3] padata

    service, host_fqdn = target_service
    sname_inner = (
        asn1_tag(0, asn1_int(2)) +                       # NT-SRV-INST
        asn1_tag(1, asn1_seq(
            asn1_gs(service.encode()) +
            asn1_gs(host_fqdn.encode())
        ))
    )

    # Etype list: RC4 (23), AES128 (17), AES256 (18)
    etypes = asn1_int(23) + asn1_int(17) + asn1_int(18)
    nonce = random.getrandbits(31)

    req_body_content = (
        asn1_tag(0, b'\x03\x05\x00\x40\x81\x00\x10') +   # KDC Options
        asn1_tag(2, asn1_gs(domcaps.encode())) +         # Realm
        asn1_tag(3, asn1_seq(sname_inner)) +             # sname
        asn1_tag(5, asn1_gt(till_ts)) +                  # Till
        asn1_tag(7, asn1_int(nonce)) +                   # Nonce
        asn1_tag(8, asn1_seq(etypes))                    # Etypes
    )
    req_body_field = asn1_tag(4, asn1_seq(req_body_content))

    # TGS-REQ
    final_content = (
        asn1_tag(1, asn1_int(5)) +                       # pvno
        asn1_tag(2, asn1_int(12)) +                      # msg-type 12=TGS-REQ
        padata_field +
        req_body_field
    )

    # Application Tag 12 (0x6c) -> TGS-REQ
    inner_seq = asn1_seq(final_content)
    return b'\x6c' + asn1_len(inner_seq) + inner_seq


def extract_ticket(data):
    pos = 0
    if data[pos] not in [0x6b, 0x6d]:
        raise ValueError("Invalid Kerberos answer")

    _, len_size = get_asn1_len(data, pos + 1)
    pos += 1 + len_size  # AS-REP

    # Skip SEQUENCE tag (0x30)
    if data[pos] != 0x30:
        raise ValueError("SEQUENCE expected")
    _, len_size = get_asn1_len(data, pos + 1)
    pos += 1 + len_size

    while pos < len(data):
        tag = data[pos]
        content_len, len_size = get_asn1_len(data, pos + 1)

        if tag == 0xa5:
            ticket_start = pos + 1 + len_size
            tkt_len, tkt_len_size = get_asn1_len(data, ticket_start + 1)
            total_tkt_size = 1 + tkt_len_size + tkt_len
            return data[ticket_start: ticket_start + total_tkt_size]

        pos += 1 + len_size + content_len

    raise ValueError("Ticket tag [5] not found in sequence")


def get_service_key(resp_bytes: bytes,
                    session_key: bytes) -> Tuple[bytes, bytes, int]:
    raw_obj, _ = decoder.decode(resp_bytes)
    enc_part = raw_obj.getComponentByPosition(5)

    cipher_blob = None
    for i in range(len(enc_part)):
        comp = enc_part.getComponentByPosition(i)
        if isinstance(comp, univ.OctetString):
            cipher_blob = bytes(comp)
            break

    if not cipher_blob:
        cipher_blob = bytes(enc_part[2])

    decrypted_raw = decrypt_kerberos_aes_cts(session_key, 8, cipher_blob)
    payload = decrypted_raw[16:]
    inner_obj, _ = decoder.decode(payload)

    key_sequence = inner_obj.getComponentByPosition(0)
    service_etype = int(key_sequence.getComponentByPosition(0))
    service_session_key = bytes(key_sequence.getComponentByPosition(1))

    ticket = raw_obj.getComponentByPosition(4)
    ticket_bytes = encoder.encode(ticket)

    if ticket_bytes[0] != 0x61:
        first_len_byte = ticket_bytes[1]
        num_len_bytes = first_len_byte & 0x7f if first_len_byte & 0x80 else 0
        ticket_bytes = ticket_bytes[2 + num_len_bytes:]

    return ticket_bytes, service_session_key, service_etype


async def get_tgs(username: str, domain: str, host: str,
                  as_rep_bytes: bytes, base_key: bytes,
                  kdc_host: str,
                  kdc_port: int = 88) -> Tuple[bytes, bytes, int]:
    tgs_session_key = get_session_key(as_rep_bytes, base_key)
    tgs_ticket = extract_ticket(as_rep_bytes)
    tgs_req = build_tgs_req(username,
                            domain,
                            tgs_session_key,
                            tgs_ticket,
                            ("host", host))
    as_res_bytes = await send_kerberos_packet(tgs_req, kdc_host, kdc_port)
    parse_krb_error(as_res_bytes)
    ticket, service_key, etype = get_service_key(as_res_bytes, tgs_session_key)
    return ticket, service_key, etype



from typing import Optional, Tuple
from datetime import datetime, timezone
from .asn1 import asn1_len, asn1_tag, asn1_gt, asn1_int, asn1_ostr
from .tools import encrypt_kerberos_rc4, decrypt_kerberos_rc4, peel_tag
from .tools import decrypt_kerberos_aes_cts, encrypt_kerberos_aes_cts
from .const import OID_SPNEGO, OID_IETF_KRB5, OID_MS_KRB5
from ..exceptions import NoNewActiveKey


def wrap_gss_kerberos(ap_req_bytes: bytes):
    outer_oid = OID_IETF_KRB5
    inner_oid = OID_MS_KRB5

    mech_types_inner = b'\x30' + asn1_len(outer_oid) + outer_oid
    mech_types = asn1_tag(0, mech_types_inner)

    inner_gss_content = inner_oid + b'\x01\x00' + ap_req_bytes
    gss_wrapper = b'\x60' + asn1_len(inner_gss_content) + inner_gss_content

    mech_token = asn1_tag(2, asn1_ostr(gss_wrapper))

    neg_body = mech_types + mech_token
    neg_init = asn1_tag(0, b'\x30' + asn1_len(neg_body) + neg_body)

    final_payload = OID_SPNEGO + neg_init

    return b'\x60' + asn1_len(final_payload) + final_payload


def build_ap_req(username: str,
                 domain: str,
                 ticket: bytes,
                 service_session_key: bytes,
                 etype: int) -> bytes:
    now = datetime.now(timezone.utc)
    timestamp = now.strftime('%Y%m%d%H%M%SZ').encode('ascii')

    gss_data = b'\x10\x00\x00\x00' + b'\x00' * 16 + b'\x3e\x10\x00\x00'
    cksum_inner = (
        asn1_tag(0, asn1_int(32771)) +
        asn1_tag(1, asn1_ostr(gss_data))
    )
    cksum_asn1 = asn1_tag(3, b'\x30\x23' + cksum_inner)

    # CName [2]
    uname = username.encode()
    uname_bytes = b'\x1b' + asn1_len(uname) + uname
    cname_inner = (
        asn1_tag(0, asn1_int(1)) +
        asn1_tag(1, b'\x30' + asn1_len(uname_bytes) + uname_bytes)
    )
    cname_asn1 = \
        asn1_tag(2, b'\x30' + asn1_len(cname_inner) + cname_inner)

    # Realm [1]
    realm = domain.encode()
    realm_asn1 = asn1_tag(1, b'\x1b' + asn1_len(realm) + realm)

    # Authenticator Body (Plaintext)
    auth_body = (
        asn1_tag(0, asn1_int(5)) +
        realm_asn1 +
        cname_asn1 +
        cksum_asn1 +
        asn1_tag(4, asn1_int(now.microsecond)) +
        asn1_tag(5, asn1_gt(timestamp)) +
        asn1_tag(7, b'\x02\x01\x00')
    )

    # Encrypt authenticator
    auth_inner = b'\x30' + asn1_len(auth_body) + auth_body
    auth_asn1 = b'\x62' + asn1_len(auth_inner) + auth_inner

    if etype == 18:  # AES-256
        enc_auth = encrypt_kerberos_aes_cts(service_session_key, 11, auth_asn1)
        etype_byte = b'\x12'
    elif etype == 17:  # AES-128
        enc_auth = encrypt_kerberos_aes_cts(service_session_key, 11, auth_asn1)
        etype_byte = b'\x11'
    elif etype == 23:  # RC4
        enc_auth = encrypt_kerberos_rc4(service_session_key, 11, auth_asn1)
        etype_byte = b'\x17'
    else:
        raise ValueError(f"Invalid E-type: {etype}")

    etype_asn1 = asn1_tag(0, b'\x02\x01' + etype_byte)
    cipher_asn1 = asn1_tag(2, asn1_ostr(enc_auth))

    enc_body = etype_asn1 + cipher_asn1
    enc_part = asn1_tag(4, b'\x30' + asn1_len(enc_body) + enc_body)

    ap_req_body = (
        asn1_tag(0, asn1_int(5)) +                      # pvno
        asn1_tag(1, asn1_int(14)) +                     # msg-type
        asn1_tag(2, b'\x03\x05\x00\x20\x00\x00\x00') +  # ap-options
        asn1_tag(3, ticket) +                           # ticket
        enc_part                                        # authenticator
    )

    inner_pdu = b'\x30' + asn1_len(ap_req_body) + ap_req_body
    return b'\x6e' + asn1_len(inner_pdu) + inner_pdu


def _rc4_helper(asn1_data: bytes) -> Tuple[Optional[bytes], Optional[int]]:
    active_key = None
    seq_number = None

    for i in range(len(asn1_data) - 10):
        if asn1_data[i] == 0xa2:
            content = peel_tag(asn1_data[i:], 0xa2)
            if content:
                val_idx = content.find(b'\xa1')
                if val_idx != -1:
                    val_wrap = peel_tag(content[val_idx:], 0xa1)
                    if val_wrap:
                        res_key = peel_tag(val_wrap, 0x04)
                        if res_key and len(res_key) == 16:
                            active_key = res_key
                            break

    for i in range(len(asn1_data) - 2):
        if asn1_data[i] == 0xa3:
            content = peel_tag(asn1_data[i:], 0xa3)
            if content:
                inner = peel_tag(content, 0x02)
                if inner:
                    seq_number = int.from_bytes(inner, 'big')
                    break
                elif 0 < len(content) <= 8:
                    seq_number = int.from_bytes(content, 'big')
                    break

    return active_key, seq_number


def _aes_helper(asn1_data: bytes) -> Tuple[Optional[bytes], Optional[int]]:
    active_key = None
    seq_number = None

    current = asn1_data
    if asn1_data.startswith(b'\x7b'):
        current = peel_tag(asn1_data, 0x7b)
        if current and current.startswith(b'\x30'):
            current = peel_tag(current, 0x30)

    if current:
        subkey_cont = peel_tag(current, 0xa2)
        if subkey_cont:
            k_seq = peel_tag(subkey_cont, 0x30)
            if k_seq:
                k_val_wrap = peel_tag(k_seq, 0xa1)
                if k_val_wrap:
                    active_key = peel_tag(k_val_wrap, 0x04)

        seq_cont = peel_tag(current, 0xa3)
        if seq_cont:
            inner_seq = peel_tag(seq_cont, 0x02)
            if inner_seq:
                seq_number = int.from_bytes(inner_seq, 'big')

    return active_key, seq_number


def get_active_key(auth_bytes: bytes,
                   service_session_key: bytes,
                   etype: int) -> Tuple[Optional[bytes], Optional[int]]:
    active_key, seq_number = None, None
    kerberos_data = auth_bytes

    if auth_bytes.startswith(b'\xa1'):
        idx_a2 = auth_bytes.find(b'\xa2')
        if idx_a2 == -1:
            raise NoNewActiveKey()

        idx_04 = auth_bytes.find(b'\x04', idx_a2)
        if idx_04 != -1:
            pos = idx_04 + 1
            lb = auth_bytes[pos]
            pos += 1
            if lb & 0x80:
                n = lb & 0x7f
                length = int.from_bytes(auth_bytes[pos:pos+n], 'big')
                pos += n
            else:
                length = lb
            kerberos_data = auth_bytes[pos: pos + length]
        else:
            raise NoNewActiveKey()

    etype_idx = kerberos_data.find(bytes([0x02, 0x01, etype]))
    if etype_idx == -1:
        return None, None

    idx_04_cipher = kerberos_data.find(b'\x04', etype_idx)
    if idx_04_cipher == -1:
        return None, None

    pos = idx_04_cipher + 1
    lb = kerberos_data[pos]
    pos += 1
    if lb & 0x80:
        n = lb & 0x7f
        c_length = int.from_bytes(kerberos_data[pos: pos + n], 'big')
        pos += n
    else:
        c_length = lb
    cipher_blob = kerberos_data[pos: pos + c_length]

    if etype == 23:  # RC4
        decrypted = \
            decrypt_kerberos_rc4(service_session_key, 12, cipher_blob)
        # Skip 8 bytes confounder
        active_key, seq_number = _rc4_helper(decrypted[8:])
    elif etype in (17, 18):  # AES
        decrypted = \
            decrypt_kerberos_aes_cts(service_session_key, 12, cipher_blob)
        # Skip 16 bytes confounder
        active_key, seq_number = _aes_helper(decrypted[16:])
    else:
        raise ValueError(f"Invalid E-type: {etype}")

    return active_key, seq_number



import datetime
import struct
from .asn1 import asn1_len, asn1_seq, asn1_tag, asn1_int, asn1_gt, asn1_ostr
from .tools import encrypt_kerberos_rc4, encrypt_kerberos_aes_cts
from ..dcom_const import NDR_TransferSyntaxIdentifier
from ..rpc.const import RPC_C_AUTHN_GSS_NEGOTIATE


def get_neg_token(service_session_key: bytes, seq_number: int, etype: int):
    now = datetime.datetime.now(datetime.timezone.utc)
    timestamp = now.strftime('%Y%m%d%H%M%SZ').encode('ascii')
    cusec = now.microsecond

    enc_ap_rep_body = (
        asn1_tag(0, asn1_gt(timestamp)) +   # [0] ctime
        asn1_tag(1, asn1_int(cusec)) +      # [1] cusec
        asn1_tag(3, asn1_int(seq_number))   # [3] seq-number
    )

    inner_seq = b'\x30' + asn1_len(enc_ap_rep_body) + enc_ap_rep_body
    plaintext = b'\x7b' + asn1_len(inner_seq) + inner_seq

    if etype in (17, 18):  # AES-128 / AES-256
        enc_data = encrypt_kerberos_aes_cts(service_session_key, 12, plaintext)
        current_etype = etype
    else:   # RC4 (23)
        enc_data = encrypt_kerberos_rc4(service_session_key, 12, plaintext)
        current_etype = 23

    enc_part_content = (
        asn1_tag(0, asn1_int(current_etype)) +
        asn1_tag(2, asn1_ostr(enc_data))
    )

    ap_rep_body = (
        asn1_tag(0, asn1_int(5)) +                   # pvno
        asn1_tag(1, asn1_int(15)) +                  # msg-type 15 (AP-REP)
        asn1_tag(2, b'\x30' + asn1_len(enc_part_content) + enc_part_content)
    )

    ap_rep_wrap = b'\x30' + asn1_len(ap_rep_body) + ap_rep_body
    ap_rep_full = b'\x6f' + asn1_len(ap_rep_wrap) + ap_rep_wrap

    # MechToken [2] -> Octet String -> AP-REP
    resp_token = asn1_tag(2, asn1_ostr(ap_rep_full))

    # NegTokenResp [1] -> Sequence -> MechToken
    auth_neg_token = asn1_tag(1, b'\x30' + asn1_len(resp_token) + resp_token)

    return auth_neg_token


def build_alter_context(iid: bytes,
                        call_id: int,
                        auth_level: int,
                        context_id: int,
                        neg_token: bytes):
    ndr_ts = NDR_TransferSyntaxIdentifier

    sec_trailer = struct.pack('<BBBB',
                              RPC_C_AUTHN_GSS_NEGOTIATE,
                              auth_level,
                              0x00, 0x00) + struct.pack('<I', context_id)

    ctx_body = (
        b'\x00\x00\x00\x00' +       # Padding
        b'\x01\x00\x00\x00' +       # NumCtxItems: 1
        b'\x00\x00' +               # Context ID: 0
        b'\x01\x00' +               # NumTransItems: 1
        iid +                       # IID (20 bytes incl version)
        ndr_ts +                    # Syntax (20 bytes)
        b'\x00\x00\x01\x00' +       # Impacket indicator / marker
        iid +                       # IID repeat (20 bytes)
        ndr_ts                      # Syntax repeat (20 bytes)
    )

    auth_len = len(neg_token)
    frag_len = 124 + auth_len

    header = (
        b'\x05\x00' +               # RPC Version 5.0
        b'\x0e' +                   # PDU Type: Alter Context
        b'\x03' +                   # Flags: First + Last
        b'\x10\x00\x00\x00' +       # Data Representation
        struct.pack('<H', frag_len) +
        struct.pack('<H', auth_len) +
        struct.pack('<I', call_id) +
        b'\xb8\x10\xb8\x10'         # Max Xmit/Recv (4280)
    )

    return header + ctx_body + sec_trailer + neg_token
